# 核心概念

- [核心概念](#%e6%a0%b8%e5%bf%83%e6%a6%82%e5%bf%b5)
  - [Index](#index)
  - [Type](#type)
  - [Document](#document)
  - [Node](#node)
  - [Cluster](#cluster)
  - [Shards and replicas](#shards-and-replicas)
    - [Shards](#shards)
    - [Replicas](#replicas)
  - [Mappings and types](#mappings-and-types)
    - [**Data types**](#data-types)
    - [Mappings](#mappings)
      - [用显式 mapping 创建索引](#%e7%94%a8%e6%98%be%e5%bc%8f-mapping-%e5%88%9b%e5%bb%ba%e7%b4%a2%e5%bc%95)
  - [Inverted index](#inverted-index)

## Index

**Index**是 ES 中存储和管理单个**Type**的文档容器

<img src="../images/pic-3.jpg" alt="关系图" width=300 height=250/>

ES 中的**Index**的概念类似关系数据库中的数据库 schema，ES 中的**Type**等效于表(table),ES 中的**Document**等效于表中的记录。但该类比只是为了便于理解，与关系型数据库架构不同，**一个 Index 只能包含一个 Type**。(6.0 开始官方不允许在一个 index 下建立多个 type，统一 type 名称：doc)

## Type

**Type**用于在**Index**中对相同类型的文档进行逻辑分组或组织

## Document

**Document**包含多个字段。JSON 文档中的每个字段都有特定类型

ES 包含一些内部元字段：

- \_id: 文档的唯一标识，类似于数据库表中的主键，它可以由用户指定或自动生成
- \_type: 该字段包含文档的类型
- \_index: 该字段包含文档的索引名

## Node

- ES 的 Node 是一个 ES 的单服务器实例，它可以是较大集群节点的一部分。它参与构建索引，搜索和执行 ES 支持的其他操作。每个 ES 的 Node 在启动时都会分配一个唯一的 ID 和名称。还可以通过 ES 配置文件`(config/elasticsearch.yml)`中的`node.name`参数为节点分配静态名称。
- 在最低的级别，一个节点对应于一个 es 进程的实例。负责管理其数据共享

## Cluster

- 一个集群承载一个或多个索引，并负责提供搜索、索引和聚合等操作。集群有一个或多个 node 组成。
- 建议修改 ES 配置文件`(config/elasticsearch.yml)`中的`cluster.name`属性，以避免加入同一个网络中的另一个集群

## Shards and replicas

### Shards

- 分片是为了解决存储大规模数据的问题，将数据切分分别存储到不同的分片中
- 将数据在分片之间划分的过程称为`sharding`
  - 它有助于利用集群不同节点进行存储
  - 它有助于利用集群不同节点的处理能力
- 默认情况下，每个索引都配置 5 个分片。在创建索引时，可以指定数据用于索引的分片数，创建索引后，无法修改分片数

下图说明了如何在三节点集群上分配一个索引的 5 个分片
<img src="../images/shard1.jpg" alt="集群节点的分片1" width=600 height=280/>

在图中，分片名为 P1 到 P5。每个分片包含大约五分之一的索引中存储的总数据。当针对此索引进行查询时，ES 会处理所有分片并合并结果

### Replicas

<img src="../images/shard2.jpg" alt="集群节点的分片2" width=600 height=280/>

当其中一个节点(节点 1)关闭，当使用到节点 1 时，我们会丢失存储在分片 P1 和 P2 中的数据共享，
尽管硬件出现故障，但 ES 依旧可以运行，它通过`副本分片`(replicas shards)或`副本`(replicas)解决,副本分片是原分片的副本，可提供高可用数据

<img src="../images/shard3.jpg" alt="集群节点的分片3" width=600 height=280/>

总之，节点(Node)聚集在一起形成一个集群。集群(Cluster)提供一个物理服务层，可以在该层创建多个索引(Index)。索引可以包含一个或多个类型(Type)，每个 type 包含数百万或数十亿个文档(Document)。索引被拆分为分片(Shard)，即索引中基础数据的分区。分片分布在集群的节点上。副本(Replicas)是主分片的副本，提供高可用性和故障转移。

## Mappings and types

### **Data types**

`ES`支持多种数据类型，以支持不同的存储场景，包含文本数据、数字、布尔值、二进制对象、数组、对象、嵌套类型、geo-points、geo-shapes 和许多其他专用数据类型(如 IPv4、IPv6 地址)

- `String` datatypes：
  - `text`：text 字段类型可用于支持包含描述或冗长文本值的字段进行全文搜索
  - `keyword`：keyword 类型支持对字符串进行分析。该类型的字段支持排序，筛选和聚合
- `Numeric` datatypes：
  - `byte，short，integer和long`：分别为带符号的 8-bit，16-bit，32-bit 和 64-bit 精度的整数
  - `float和double`：具有单精度 32 位和双精度 64 位的 IEEE 754 浮点数
  - `half_float`：具有半精度 16 位表示的浮点数
  - `scaled_float`：缩放类型的的浮点数，由一个 long 型和一个缩放因子(scaling_factor)一起组成。比如价格只需要精确到分，price 为 57.34 的字段缩放因子为 100，存起来就是 5734。(`A finite floating point number that is backed by a long, scaled by a fixed double scaling factor`.)
- `Date` datatype：
  - `date`：具有可选时间戳组件的日期，该组件能够将精确时间戳存储到毫秒
- `Boolean datatype`：
  - `boolean`：boolean
- `Binary datatype`：
  - `binary`：允许执行 Base64 编码后存储二进制值
- `Range datatypes`：
  - `integer_range`, `float_range`, `long_range`, `double_range` and `date_range`：定义整型，浮点型，长整型和双精度和日期等范围
- `Complex datatypes`：
  - `Array datatype`：具有相同类型元素的数组。例如，字符串数组、整数数组等。不允许在数组中混合数据类型
  - `Object datatype`：允许 JSON 文档中的内部对象
  - `Nested datatype`：可用于支持内部对象的数组(对象数组)
- `Other datatypes`：
  - `Geo-point datatype`：允许将地理点存储为经度和纬度。
  - `Geo-Shape datatype`：`Geo-Shape`数据类型有助于使用任意地理形状（如矩形和多边形）进行索引和搜索。当被索引的数据或正在执行的查询包含除了点之外的形状时，应该使用它。
  - `IP datadype`：允许存储 IPv4 和 IPv6 的地址

### Mappings

```json
PUT /catalog/product/2
{
  "sku": "SP000002",
  "title": "Google Pixel Phone 32GB - 5 inch display",
  "description": "Google Pixel Phone 32GB - 5 inch display (Factory Unlocked US Version)",
  "price": 400.00,
  "resolution": "1440 x 2560 pixels",
  "os": "Android 7.1"
}
```

#### 用显式 mapping 创建索引

```json
# 创建
PUT /my-index
{
  "mappings": {
    "properties": {
      "age": { "type": "integer" },
      "email": { "type": "keyword"  },
      "name": { "type": "text"  }
    }
  }
}

# 添加
PUT /my-index/_mapping
{
  "properties": {
    "employee-id": {
      "type": "keyword",
      "index": false
    }
  }
}
```

## Inverted index

| Document ID | Document                            |
| ----------- | ----------------------------------- |
| 1           | It is Sunday tomorrow.              |
| 2           | Sunday is the last day of the week. |
| 3           | The choice is yours.                |

| Term     | Frequency | Documents(Postings) |
| -------- | --------- | ------------------- |
| choice   | 1         | 3                   |
| day      | 1         | 2                   |
| is       | 3         | 1,2,3               |
| it       | 1         | 1                   |
| last     | 1         | 2                   |
| of       | 1         | 2                   |
| sunday   | 2         | 1,2                 |
| the      | 3         | 2,3                 |
| tomorrow | 1         | 1                   |
| week     | 1         | 2                   |
| yours    | 1         | 3                   |

* 注意
  * 文档(`Document`)在删除标点符号并转换为小写后被分解为项(`term`)
  * 项以字母顺序排序
  * `Frequency`列展示了`term`在整个文档集中的显示次数
  * `Documents(Postings)`显示`term`在哪个文档出现，此外还可以包含找到该`term`的确切位置(文档中的`offset`)

[文章参考-ElasticSearch 6.x 初体验](https://juejin.im/post/5b799bff51882542dd224943)
